@@ -1,19 +1,19 @@
  /*
- * Copyright (c) 2014 Yubico AB
+ * Copyright (c) 2014-2016 Yubico AB
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
  * met:
- * 
+ *
  *   * Redistributions of source code must retain the above copyright
  *     notice, this list of conditions and the following disclaimer.
- * 
+ *
  *   * Redistributions in binary form must reproduce the above
  *     copyright notice, this list of conditions and the following
  *     disclaimer in the documentation and/or other materials provided
  *     with the distribution.
- * 
+ *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
  * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
  * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
@@ -103,6 +103,7 @@ ykpiv_rc ykpiv_init(ykpiv_state **state, int verbose) {
   }
   memset(s, 0, sizeof(ykpiv_state));
   s->verbose = verbose;
+  s->context = SCARD_E_INVALID_HANDLE;
   *state = s;
   return YKPIV_OK;
 }
@@ -119,50 +120,24 @@ ykpiv_rc ykpiv_disconnect(ykpiv_state *state) {
     state->card = 0;
   }
 
-  if(state->context) {
+  if(SCardIsValidContext(state->context) == SCARD_S_SUCCESS) {
     SCardReleaseContext(state->context);
-    state->context = 0;
+    state->context = SCARD_E_INVALID_HANDLE;
   }
 
   return YKPIV_OK;
 }
 
 ykpiv_rc ykpiv_connect(ykpiv_state *state, const char *wanted) {
-  unsigned long num_readers = 0;
   unsigned long active_protocol;
-  char reader_buf[1024];
+  char reader_buf[2048];
+  size_t num_readers = sizeof(reader_buf);
   long rc;
   char *reader_ptr;
 
-  rc = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &state->context);
-  if (rc != SCARD_S_SUCCESS) {
-    if(state->verbose) {
-      fprintf (stderr, "error: SCardEstablishContext failed, rc=%08lx\n", rc);
-    }
-    return YKPIV_PCSC_ERROR;
-  }
-
-  rc = SCardListReaders(state->context, NULL, NULL, &num_readers);
-  if (rc != SCARD_S_SUCCESS) {
-    if(state->verbose) {
-      fprintf (stderr, "error: SCardListReaders failed, rc=%08lx\n", rc);
-    }
-    SCardReleaseContext(state->context);
-    return YKPIV_PCSC_ERROR;
-  }
-
-  if (num_readers > sizeof(reader_buf)) {
-    num_readers = sizeof(reader_buf);
-  }
-
-  rc = SCardListReaders(state->context, NULL, reader_buf, &num_readers);
-  if (rc != SCARD_S_SUCCESS)
-  {
-    if(state->verbose) {
-      fprintf (stderr, "error: SCardListReaders failed, rc=%08lx\n", rc);
-    }
-    SCardReleaseContext(state->context);
-    return YKPIV_PCSC_ERROR;
+  ykpiv_rc ret = ykpiv_list_readers(state, reader_buf, &num_readers);
+  if(ret != YKPIV_OK) {
+    return ret;
   }
 
   for(reader_ptr = reader_buf; *reader_ptr != '\0'; reader_ptr += strlen(reader_ptr) + 1) {
@@ -205,11 +180,11 @@ ykpiv_rc ykpiv_connect(ykpiv_state *state, const char *wanted) {
           fprintf(stderr, "Failed communicating with card: '%s'\n", ykpiv_strerror(res));
         }
         continue;
-      } else if(sw == 0x9000) {
+      } else if(sw == SW_SUCCESS) {
         return YKPIV_OK;
       } else {
         if(state->verbose) {
-          fprintf(stderr, "Failed selecting applet: %04x\n", sw);
+          fprintf(stderr, "Failed selecting application: %04x\n", sw);
         }
       }
     }
@@ -217,15 +192,60 @@ ykpiv_rc ykpiv_connect(ykpiv_state *state, const char *wanted) {
 
   if(*reader_ptr == '\0') {
     if(state->verbose) {
-      fprintf(stderr, "error: no useable reader found.\n");
+      fprintf(stderr, "error: no usable reader found.\n");
     }
     SCardReleaseContext(state->context);
+    state->context = SCARD_E_INVALID_HANDLE;
     return YKPIV_PCSC_ERROR;
   }
 
   return YKPIV_GENERIC_ERROR;
 }
 
+ykpiv_rc ykpiv_list_readers(ykpiv_state *state, char *readers, size_t *len) {
+  unsigned long num_readers = 0;
+  long rc;
+
+  if(SCardIsValidContext(state->context) != SCARD_S_SUCCESS) {
+    rc = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &state->context);
+    if (rc != SCARD_S_SUCCESS) {
+      if(state->verbose) {
+        fprintf (stderr, "error: SCardEstablishContext failed, rc=%08lx\n", rc);
+      }
+      return YKPIV_PCSC_ERROR;
+    }
+  }
+
+  rc = SCardListReaders(state->context, NULL, NULL, &num_readers);
+  if (rc != SCARD_S_SUCCESS) {
+    if(state->verbose) {
+      fprintf (stderr, "error: SCardListReaders failed, rc=%08lx\n", rc);
+    }
+    SCardReleaseContext(state->context);
+    state->context = SCARD_E_INVALID_HANDLE;
+    return YKPIV_PCSC_ERROR;
+  }
+
+  if (num_readers > *len) {
+    num_readers = *len;
+  }
+
+  rc = SCardListReaders(state->context, NULL, readers, &num_readers);
+  if (rc != SCARD_S_SUCCESS)
+  {
+    if(state->verbose) {
+      fprintf (stderr, "error: SCardListReaders failed, rc=%08lx\n", rc);
+    }
+    SCardReleaseContext(state->context);
+    state->context = SCARD_E_INVALID_HANDLE;
+    return YKPIV_PCSC_ERROR;
+  }
+
+  *len = num_readers;
+
+  return YKPIV_OK;
+}
+
 ykpiv_rc ykpiv_transfer_data(ykpiv_state *state, const unsigned char *templ,
     const unsigned char *in_data, long in_len,
     unsigned char *out_data, unsigned long *out_len, int *sw) {
@@ -238,7 +258,7 @@ ykpiv_rc ykpiv_transfer_data(ykpiv_state *state, const unsigned char *templ,
   rc = SCardBeginTransaction(state->card);
   if(rc != SCARD_S_SUCCESS) {
     if(state->verbose) {
-      fprintf(stderr, "error: Failed to being pcsc transaction, rc=%08lx\n", rc);
+      fprintf(stderr, "error: Failed to begin pcsc transaction, rc=%08lx\n", rc);
     }
     return YKPIV_PCSC_ERROR;
   }
@@ -263,12 +283,12 @@ ykpiv_rc ykpiv_transfer_data(ykpiv_state *state, const unsigned char *templ,
     res = send_data(state, &apdu, data, &recv_len, sw);
     if(res != YKPIV_OK) {
       return res;
-    } else if(*sw != 0x9000 && *sw >> 8 != 0x61) {
+    } else if(*sw != SW_SUCCESS && *sw >> 8 != 0x61) {
       return YKPIV_OK;
     }
     if(*out_len + recv_len - 2 > max_out) {
       if(state->verbose) {
-	fprintf(stderr, "Output buffer to small, wanted to write %lu, max was %lu.\n", *out_len + recv_len - 2, max_out);
+  fprintf(stderr, "Output buffer to small, wanted to write %lu, max was %lu.\n", *out_len + recv_len - 2, max_out);
       }
       return YKPIV_SIZE_ERROR;
     }
@@ -293,7 +313,7 @@ ykpiv_rc ykpiv_transfer_data(ykpiv_state *state, const unsigned char *templ,
     res = send_data(state, &apdu, data, &recv_len, sw);
     if(res != YKPIV_OK) {
       return res;
-    } else if(*sw != 0x9000 && *sw >> 8 != 0x61) {
+    } else if(*sw != SW_SUCCESS && *sw >> 8 != 0x61) {
       return YKPIV_OK;
     }
     if(*out_len + recv_len - 2 > max_out) {
@@ -370,16 +390,16 @@ ykpiv_rc ykpiv_authenticate(ykpiv_state *state, unsigned const char *key) {
   /* get a challenge from the card */
   {
     memset(apdu.raw, 0, sizeof(apdu));
-    apdu.st.ins = 0x87;
-    apdu.st.p1 = 0x03; /* triple des */
-    apdu.st.p2 = 0x9b; /* management key */
+    apdu.st.ins = YKPIV_INS_AUTHENTICATE;
+    apdu.st.p1 = YKPIV_ALGO_3DES; /* triple des */
+    apdu.st.p2 = YKPIV_KEY_CARDMGM; /* management key */
     apdu.st.lc = 0x04;
     apdu.st.data[0] = 0x7c;
     apdu.st.data[1] = 0x02;
     apdu.st.data[2] = 0x80;
     if((res = send_data(state, &apdu, data, &recv_len, &sw)) != YKPIV_OK) {
       return res;
-    } else if(sw != 0x9000) {
+    } else if(sw != SW_SUCCESS) {
       return YKPIV_AUTHENTICATION_ERROR;
     }
     memcpy(challenge, data + 4, 8);
@@ -406,7 +426,7 @@ ykpiv_rc ykpiv_authenticate(ykpiv_state *state, unsigned const char *key) {
     *dataptr++ = 8;
     if(RAND_pseudo_bytes(dataptr, 8) == -1) {
       if(state->verbose) {
-	fprintf(stderr, "Failed getting randomness for authentication.\n");
+  fprintf(stderr, "Failed getting randomness for authentication.\n");
       }
       return YKPIV_RANDOMNESS_ERROR;
     }
@@ -415,7 +435,7 @@ ykpiv_rc ykpiv_authenticate(ykpiv_state *state, unsigned const char *key) {
     apdu.st.lc = dataptr - apdu.st.data;
     if((res = send_data(state, &apdu, data, &recv_len, &sw)) != YKPIV_OK) {
       return res;
-    } else if(sw != 0x9000) {
+    } else if(sw != SW_SUCCESS) {
       return YKPIV_AUTHENTICATION_ERROR;
     }
   }
@@ -433,6 +453,10 @@ ykpiv_rc ykpiv_authenticate(ykpiv_state *state, unsigned const char *key) {
 }
 
 ykpiv_rc ykpiv_set_mgmkey(ykpiv_state *state, const unsigned char *new_key) {
+  return ykpiv_set_mgmkey2(state, new_key, 0);
+}
+
+ykpiv_rc ykpiv_set_mgmkey2(ykpiv_state *state, const unsigned char *new_key, const unsigned char touch) {
   APDU apdu;
   unsigned char data[261];
   unsigned long recv_len = sizeof(data);
@@ -446,11 +470,11 @@ ykpiv_rc ykpiv_set_mgmkey(ykpiv_state *state, const unsigned char *new_key) {
     DES_set_odd_parity(&key_tmp);
     if(DES_is_weak_key(&key_tmp) != 0) {
       if(state->verbose) {
-	fprintf(stderr, "Won't set new key '");
-	dump_hex(new_key + i * 8, 8);
-	fprintf(stderr, "' since it's weak (with parity the key is: ");
-	dump_hex(key_tmp, 8);
-	fprintf(stderr, ").\n");
+  fprintf(stderr, "Won't set new key '");
+  dump_hex(new_key + i * 8, 8);
+  fprintf(stderr, "' since it's weak (with parity the key is: ");
+  dump_hex(key_tmp, 8);
+  fprintf(stderr, ").\n");
       }
       return YKPIV_GENERIC_ERROR;
     }
@@ -459,7 +483,13 @@ ykpiv_rc ykpiv_set_mgmkey(ykpiv_state *state, const unsigned char *new_key) {
   memset(apdu.raw, 0, sizeof(apdu));
   apdu.st.ins = YKPIV_INS_SET_MGMKEY;
   apdu.st.p1 = 0xff;
-  apdu.st.p2 = 0xff;
+  if(touch == 0) {
+    apdu.st.p2 = 0xff;
+  } else if(touch == 1) {
+    apdu.st.p2 = 0xfe;
+  } else {
+    return YKPIV_GENERIC_ERROR;
+  }
   apdu.st.lc = DES_KEY_SZ * 3 + 3;
   apdu.st.data[0] = YKPIV_ALGO_3DES;
   apdu.st.data[1] = YKPIV_KEY_CARDMGM;
@@ -467,7 +497,7 @@ ykpiv_rc ykpiv_set_mgmkey(ykpiv_state *state, const unsigned char *new_key) {
   memcpy(apdu.st.data + 3, new_key, DES_KEY_SZ * 3);
   if((res = send_data(state, &apdu, data, &recv_len, &sw)) != YKPIV_OK) {
     return res;
-  } else if(sw == 0x9000) {
+  } else if(sw == SW_SUCCESS) {
     return YKPIV_OK;
   }
   return YKPIV_GENERIC_ERROR;
@@ -505,7 +535,7 @@ ykpiv_rc ykpiv_hex_decode(const char *hex_in, size_t in_len,
 }
 
 static ykpiv_rc _general_authenticate(ykpiv_state *state,
-    const unsigned char *raw_in, size_t in_len,
+    const unsigned char *sign_in, size_t in_len,
     unsigned char *out, size_t *out_len,
     unsigned char algorithm, unsigned char key, bool decipher) {
   unsigned char indata[1024];
@@ -513,8 +543,7 @@ static ykpiv_rc _general_authenticate(ykpiv_state *state,
   unsigned char data[1024];
   unsigned char templ[] = {0, YKPIV_INS_AUTHENTICATE, algorithm, key};
   unsigned long recv_len = sizeof(data);
-  unsigned char sign_in[256];
-  size_t pad_len = 0;
+  size_t key_len = 0;
   int sw;
   size_t bytes;
   size_t len = 0;
@@ -522,31 +551,26 @@ static ykpiv_rc _general_authenticate(ykpiv_state *state,
 
   switch(algorithm) {
     case YKPIV_ALGO_RSA1024:
-      pad_len = 128;
+      key_len = 128;
     case YKPIV_ALGO_RSA2048:
-      if(pad_len == 0) {
-	pad_len = 256;
+      if(key_len == 0) {
+	key_len = 256;
       }
-      if(!decipher) {
-	if(in_len + RSA_PKCS1_PADDING_SIZE > pad_len) {
-	  return YKPIV_SIZE_ERROR;
-	}
-	RSA_padding_add_PKCS1_type_1(sign_in, pad_len, raw_in, in_len);
-	in_len = pad_len;
-      } else {
-	if(in_len != pad_len) {
-	  return YKPIV_SIZE_ERROR;
-	}
-	memcpy(sign_in, raw_in, in_len);
+      if(in_len != key_len) {
+	return YKPIV_SIZE_ERROR;
       }
       break;
     case YKPIV_ALGO_ECCP256:
-      if(!decipher && in_len > 32) {
+      key_len = 32;
+    case YKPIV_ALGO_ECCP384:
+      if(key_len == 0) {
+	key_len = 48;
+      }
+      if(!decipher && in_len > key_len) {
 	return YKPIV_SIZE_ERROR;
-      } else if(decipher && in_len != 65) {
+      } else if(decipher && in_len != (key_len * 2) + 1) {
 	return YKPIV_SIZE_ERROR;
       }
-      memcpy(sign_in, raw_in, in_len);
       break;
     default:
       return YKPIV_ALGORITHM_ERROR;
@@ -564,7 +588,7 @@ static ykpiv_rc _general_authenticate(ykpiv_state *state,
   dataptr += set_length(dataptr, in_len + bytes + 3);
   *dataptr++ = 0x82;
   *dataptr++ = 0x00;
-  *dataptr++ = algorithm == YKPIV_ALGO_ECCP256 && decipher ? 0x85 : 0x81;
+  *dataptr++ = YKPIV_IS_EC(algorithm) && decipher ? 0x85 : 0x81;
   dataptr += set_length(dataptr, in_len);
   memcpy(dataptr, sign_in, (size_t)in_len);
   dataptr += in_len;
@@ -575,11 +599,14 @@ static ykpiv_rc _general_authenticate(ykpiv_state *state,
       fprintf(stderr, "Sign command failed to communicate.\n");
     }
     return res;
-  } else if(sw != 0x9000) {
+  } else if(sw != SW_SUCCESS) {
     if(state->verbose) {
       fprintf(stderr, "Failed sign command with code %x.\n", sw);
     }
-    return YKPIV_GENERIC_ERROR;
+    if (sw == SW_ERR_SECURITY_STATUS)
+      return YKPIV_AUTHENTICATION_ERROR;
+    else
+      return YKPIV_GENERIC_ERROR;
   }
   /* skip the first 7c tag */
   if(data[0] != 0x7c) {
@@ -616,15 +643,14 @@ ykpiv_rc ykpiv_sign_data(ykpiv_state *state,
     unsigned char algorithm, unsigned char key) {
 
   return _general_authenticate(state, raw_in, in_len, sign_out, out_len,
-      algorithm, key, false);
+                               algorithm, key, false);
 }
 
-
 ykpiv_rc ykpiv_decipher_data(ykpiv_state *state, const unsigned char *in,
     size_t in_len, unsigned char *out, size_t *out_len,
     unsigned char algorithm, unsigned char key) {
   return _general_authenticate(state, in, in_len, out, out_len,
-      algorithm, key, true);
+                               algorithm, key, true);
 }
 
 ykpiv_rc ykpiv_get_version(ykpiv_state *state, char *version, size_t len) {
@@ -638,7 +664,7 @@ ykpiv_rc ykpiv_get_version(ykpiv_state *state, char *version, size_t len) {
   apdu.st.ins = YKPIV_INS_GET_VERSION;
   if((res = send_data(state, &apdu, data, &recv_len, &sw)) != YKPIV_OK) {
     return res;
-  } else if(sw == 0x9000) {
+  } else if(sw == SW_SUCCESS) {
     int result = snprintf(version, len, "%d.%d.%d", data[0], data[1], data[2]);
     if(result < 0) {
       return YKPIV_SIZE_ERROR;
@@ -677,12 +703,12 @@ ykpiv_rc ykpiv_verify(ykpiv_state *state, const char *pin, int *tries) {
   }
   if((res = send_data(state, &apdu, data, &recv_len, &sw)) != YKPIV_OK) {
     return res;
-  } else if(sw == 0x9000) {
+  } else if(sw == SW_SUCCESS) {
     return YKPIV_OK;
   } else if((sw >> 8) == 0x63) {
     *tries = (sw & 0xf);
     return YKPIV_WRONG_PIN;
-  } else if(sw == 0x6983) {
+  } else if(sw == SW_ERR_AUTH_BLOCKED) {
     *tries = 0;
     return YKPIV_WRONG_PIN;
   } else {
@@ -690,6 +716,68 @@ ykpiv_rc ykpiv_verify(ykpiv_state *state, const char *pin, int *tries) {
   }
 }
 
+#define CHREF_ACT_CHANGE_PIN 0
+#define CHREF_ACT_UNBLOCK_PIN 1
+#define CHREF_ACT_CHANGE_PUK 2
+
+static ykpiv_rc _change_pin_internal(ykpiv_state *state, int action, const char * current_pin, size_t current_pin_len, const char * new_pin, size_t new_pin_len, int *tries) {
+  int sw;
+  unsigned char templ[] = {0, YKPIV_INS_CHANGE_REFERENCE, 0, 0x80};
+  unsigned char indata[0x10];
+  unsigned char data[0xff];
+  unsigned long recv_len = sizeof(data);
+  ykpiv_rc res;
+  if (current_pin_len > 8) {
+    return YKPIV_SIZE_ERROR;
+  }
+  if (new_pin_len > 8) {
+    return YKPIV_SIZE_ERROR;
+  }
+  if(action == CHREF_ACT_UNBLOCK_PIN) {
+    templ[1] = YKPIV_INS_RESET_RETRY;
+  }
+  else if(action == CHREF_ACT_CHANGE_PUK) {
+    templ[3] = 0x81;
+  }
+  memcpy(indata, current_pin, current_pin_len);
+  if(current_pin_len < 8) {
+    memset(indata + current_pin_len, 0xff, 8 - current_pin_len);
+  }
+  memcpy(indata + 8, new_pin, new_pin_len);
+  if(new_pin_len < 8) {
+    memset(indata + 8 + new_pin_len, 0xff, 8 - new_pin_len);
+  }
+  res = ykpiv_transfer_data(state, templ, indata, sizeof(indata), data, &recv_len, &sw);
+  if(res != YKPIV_OK) {
+    return res;
+  } else if(sw != SW_SUCCESS) {
+    if((sw >> 8) == 0x63) {
+      *tries = sw & 0xf;
+      return YKPIV_WRONG_PIN;
+    } else if(sw == SW_ERR_AUTH_BLOCKED) {
+      return YKPIV_PIN_LOCKED;
+    } else {
+      if(state->verbose) {
+        fprintf(stderr, "Failed changing pin, token response code: %x.\n", sw);
+      }
+      return YKPIV_GENERIC_ERROR;
+    }
+  }
+  return YKPIV_OK;
+}
+
+ykpiv_rc ykpiv_change_pin(ykpiv_state *state, const char * current_pin, size_t current_pin_len, const char * new_pin, size_t new_pin_len, int *tries) {
+  return _change_pin_internal(state, CHREF_ACT_CHANGE_PIN, current_pin, current_pin_len, new_pin, new_pin_len, tries);
+}
+
+ykpiv_rc ykpiv_change_puk(ykpiv_state *state, const char * current_puk, size_t current_puk_len, const char * new_puk, size_t new_puk_len, int *tries) {
+  return _change_pin_internal(state, CHREF_ACT_CHANGE_PUK, current_puk, current_puk_len, new_puk, new_puk_len, tries);
+}
+
+ykpiv_rc ykpiv_unblock_pin(ykpiv_state *state, const char * puk, size_t puk_len, const char * new_pin, size_t new_pin_len, int *tries) {
+  return _change_pin_internal(state, CHREF_ACT_UNBLOCK_PIN, puk, puk_len, new_pin, new_pin_len, tries);
+}
+
 ykpiv_rc ykpiv_fetch_object(ykpiv_state *state, int object_id,
     unsigned char *data, unsigned long *len) {
   int sw;
@@ -708,7 +796,7 @@ ykpiv_rc ykpiv_fetch_object(ykpiv_state *state, int object_id,
     return res;
   }
 
-  if(sw == 0x9000) {
+  if(sw == SW_SUCCESS) {
     size_t outlen;
     int offs = get_length(data + 1, &outlen);
     if(offs == 0) {
@@ -725,7 +813,7 @@ ykpiv_rc ykpiv_fetch_object(ykpiv_state *state, int object_id,
 ykpiv_rc ykpiv_save_object(ykpiv_state *state, int object_id,
     unsigned char *indata, size_t len) {
 
-  unsigned char data[2048];
+  unsigned char data[3072];
   unsigned char *dataptr = data;
   unsigned char templ[] = {0, YKPIV_INS_PUT_DATA, 0x3f, 0xff};
   int sw;
@@ -745,13 +833,135 @@ ykpiv_rc ykpiv_save_object(ykpiv_state *state, int object_id,
   dataptr += len;
 
   if((res = ykpiv_transfer_data(state, templ, data, dataptr - data, NULL, &outlen,
-	  &sw)) != YKPIV_OK) {
+    &sw)) != YKPIV_OK) {
     return res;
   }
 
-  if(sw == 0x9000) {
+  if(sw == SW_SUCCESS) {
     return YKPIV_OK;
   } else {
     return YKPIV_GENERIC_ERROR;
   }
 }
+
+ykpiv_rc ykpiv_import_private_key(ykpiv_state *state, const unsigned char key, unsigned char algorithm,
+                                  const unsigned char *p, size_t p_len,
+                                  const unsigned char *q, size_t q_len,
+                                  const unsigned char *dp, size_t dp_len,
+                                  const unsigned char *dq, size_t dq_len,
+                                  const unsigned char *qinv, size_t qinv_len,
+                                  const unsigned char *ec_data, unsigned char ec_data_len,
+                                  const unsigned char pin_policy, const unsigned char touch_policy) {
+
+  unsigned char key_data[1024];
+  unsigned char *in_ptr = key_data;
+  unsigned char templ[] = {0, YKPIV_INS_IMPORT_KEY, algorithm, key};
+  unsigned char data[256];
+  unsigned long recv_len = sizeof(data);
+  unsigned elem_len;
+  int sw;
+  const unsigned char *params[5];
+  size_t lens[5];
+  size_t padding;
+  unsigned char n_params;
+  int i;
+  int param_tag;
+
+  if (state == NULL)
+    return YKPIV_GENERIC_ERROR;
+
+  if (key == YKPIV_KEY_CARDMGM ||
+      key < YKPIV_KEY_RETIRED1 ||
+      (key > YKPIV_KEY_RETIRED20 && key < YKPIV_KEY_AUTHENTICATION) ||
+      (key > YKPIV_KEY_CARDAUTH && key != YKPIV_KEY_ATTESTATION)) {
+    return YKPIV_KEY_ERROR;
+  }
+
+  if (pin_policy != YKPIV_PINPOLICY_DEFAULT &&
+      pin_policy != YKPIV_PINPOLICY_NEVER &&
+      pin_policy != YKPIV_PINPOLICY_ONCE &&
+      pin_policy != YKPIV_PINPOLICY_ALWAYS)
+    return YKPIV_GENERIC_ERROR;
+
+  if (touch_policy != YKPIV_TOUCHPOLICY_DEFAULT &&
+      touch_policy != YKPIV_TOUCHPOLICY_NEVER &&
+      touch_policy != YKPIV_TOUCHPOLICY_ALWAYS &&
+      touch_policy != YKPIV_TOUCHPOLICY_CACHED)
+    return YKPIV_GENERIC_ERROR;
+
+  if (algorithm == YKPIV_ALGO_RSA1024 || algorithm == YKPIV_ALGO_RSA2048) {
+
+    if (algorithm == YKPIV_ALGO_RSA1024)
+      elem_len = 64;
+    if (algorithm == YKPIV_ALGO_RSA2048)
+      elem_len = 128;
+
+    if (p == NULL || q == NULL || dp == NULL ||
+        dq == NULL || qinv == NULL)
+      return YKPIV_GENERIC_ERROR;
+
+    params[0] = p;
+    lens[0] = p_len;
+    params[1] = q;
+    lens[1] = q_len;
+    params[2] = dp;
+    lens[2] = dp_len;
+    params[3] = dq;
+    lens[3] = dq_len;
+    params[4] = qinv;
+    lens[4] = qinv_len;
+    param_tag = 0x01;
+
+    n_params = 5;
+  }
+  else if (algorithm == YKPIV_ALGO_ECCP256 || algorithm == YKPIV_ALGO_ECCP384) {
+    if (algorithm == YKPIV_ALGO_ECCP256)
+      elem_len = 32;
+    if (algorithm == YKPIV_ALGO_ECCP384)
+      elem_len = 48;
+
+    if (ec_data == NULL)
+      return YKPIV_GENERIC_ERROR;
+
+    params[0] = ec_data;
+    lens[0] = ec_data_len;
+    param_tag = 0x06;
+    n_params = 1;
+  }
+  else
+    return YKPIV_ALGORITHM_ERROR;
+
+  for (i = 0; i < n_params; i++) {
+    *in_ptr++ = param_tag + i;
+    in_ptr += set_length(in_ptr, elem_len);
+    padding = elem_len - lens[i];
+    memset(in_ptr, 0, padding);
+    in_ptr += padding;
+    memcpy(in_ptr, params[i], lens[i]);
+    in_ptr += lens[i];
+  }
+
+  if (pin_policy != YKPIV_PINPOLICY_DEFAULT) {
+    *in_ptr++ = YKPIV_PINPOLICY_TAG;
+    *in_ptr++ = 0x01;
+    *in_ptr++ = pin_policy;
+  }
+
+  if (touch_policy != YKPIV_TOUCHPOLICY_DEFAULT) {
+    *in_ptr++ = YKPIV_TOUCHPOLICY_TAG;
+    *in_ptr++ = 0x01;
+    *in_ptr++ = touch_policy;
+  }
+
+  if (ykpiv_transfer_data(state, templ, key_data, in_ptr - key_data, data, &recv_len, &sw) != YKPIV_OK)
+    return YKPIV_GENERIC_ERROR;
+
+  if (sw == SW_ERR_SECURITY_STATUS)
+    return YKPIV_AUTHENTICATION_ERROR;
+
+  if (sw != SW_SUCCESS)
+    return YKPIV_GENERIC_ERROR;
+
+  return YKPIV_OK;
+
+}
